@@ -1,6 +1,6 @@
 #!/usr/bin/perl -w
 
-#use lib '../blib/lib', '../blib/arch';
+use lib '../blib/lib', '../blib/arch';
 
 use strict;
 use Algorithm::DecisionTree;
@@ -9,8 +9,9 @@ my $parameter_file = "param.txt";
 my $output_data_file = "training.dat";
 
 my $data_gen = Algorithm::DecisionTree->training_data_generator( 
-                              output_datafile => $output_data_file,
+                              output_datafile   => $output_data_file,
                               parameter_file    => $parameter_file,
+                              write_to_file     => 1,
                               number_of_training_samples => 35,
     );
 
